@@ -7,12 +7,14 @@
 #include <gmp.h>
 #else
 #include <mini-gmp.h>
+#include <stdarg.h>
 #include <stdio.h>
 /* mini-gmp doesn't come with gmp_vfprintf, so we use our own minimal variant */
 extern int mpz_vfprintf(FILE *fp, const char *format, va_list args);
 #define gmp_vfprintf mpz_vfprintf
 #endif
 
+#include <inttypes.h>
 #include <asm/byteorder.h>
 
 enum mpz_word_order {
