@@ -64,6 +64,18 @@ static const struct symbol_table ct_dir_tbl = {
 	}
 };
 
+const char *ct_dir2str(int dir)
+{
+	const struct symbolic_constant *s;
+
+	for (s = ct_dir_tbl.symbols; s->identifier != NULL; s++) {
+		if (dir == (int)s->value)
+			return s->identifier;
+	}
+
+	return NULL;
+}
+
 const struct datatype ct_dir_type = {
 	.type		= TYPE_CT_DIR,
 	.name		= "ct_dir",
@@ -133,20 +145,30 @@ static struct symbol_table *ct_label_tbl;
 
 #define CT_LABEL_BIT_SIZE 128
 
+const char *ct_label2str(unsigned long value)
+{
+	const struct symbolic_constant *s;
+
+	for (s = ct_label_tbl->symbols; s->identifier; s++) {
+		if (value == s->value)
+			return s->identifier;
+	}
+
+	return NULL;
+}
+
 static void ct_label_type_print(const struct expr *expr,
 				 struct output_ctx *octx)
 {
 	unsigned long bit = mpz_scan1(expr->value, 0);
-	const struct symbolic_constant *s;
+	const char *labelstr = ct_label2str(bit);
 
-	for (s = ct_label_tbl->symbols; s->identifier != NULL; s++) {
-		if (bit != s->value)
-			continue;
-		nft_print(octx, "\"%s\"", s->identifier);
+	if (labelstr) {
+		nft_print(octx, "\"%s\"", labelstr);
 		return;
 	}
 	/* can happen when connlabel.conf is altered after rules were added */
-	nft_print(octx, "%ld", (long)mpz_scan1(expr->value, 0));
+	nft_print(octx, "%lu", bit);
 }
 
 static struct error_record *ct_label_type_parse(const struct expr *sym,
@@ -203,6 +225,7 @@ static const struct datatype ct_label_type = {
 	.size		= CT_LABEL_BIT_SIZE,
 	.basetype	= &bitmask_type,
 	.print		= ct_label_type_print,
+	.json		= ct_label_type_json,
 	.parse		= ct_label_type_parse,
 };
 
@@ -220,7 +243,7 @@ void ct_label_table_exit(void)
 #define NF_CT_HELPER_NAME_LEN	16
 #endif
 
-static const struct ct_template ct_templates[] = {
+const struct ct_template ct_templates[__NFT_CT_MAX] = {
 	[NFT_CT_STATE]		= CT_TEMPLATE("state",	    &ct_state_type,
 					      BYTEORDER_HOST_ENDIAN,
 					      4 * BITS_PER_BYTE),
@@ -268,24 +291,28 @@ static const struct ct_template ct_templates[] = {
 					      BYTEORDER_HOST_ENDIAN, 16),
 	[NFT_CT_EVENTMASK]	= CT_TEMPLATE("event", &ct_event_type,
 					      BYTEORDER_HOST_ENDIAN, 32),
+	[NFT_CT_SRC_IP]		= CT_TEMPLATE("ip saddr", &ipaddr_type,
+					      BYTEORDER_BIG_ENDIAN, 0),
+	[NFT_CT_DST_IP]		= CT_TEMPLATE("ip daddr", &ipaddr_type,
+					      BYTEORDER_BIG_ENDIAN, 0),
+	[NFT_CT_SRC_IP6]	= CT_TEMPLATE("ip6 saddr", &ip6addr_type,
+					      BYTEORDER_BIG_ENDIAN, 0),
+	[NFT_CT_DST_IP6]	= CT_TEMPLATE("ip6 daddr", &ip6addr_type,
+					      BYTEORDER_BIG_ENDIAN, 0),
 };
 
 static void ct_print(enum nft_ct_keys key, int8_t dir, uint8_t nfproto,
 		     struct output_ctx *octx)
 {
-	const struct symbolic_constant *s;
+	const char *dirstr = ct_dir2str(dir);
 	const struct proto_desc *desc;
 
 	nft_print(octx, "ct ");
 	if (dir < 0)
 		goto done;
 
-	for (s = ct_dir_tbl.symbols; s->identifier != NULL; s++) {
-		if (dir == (int)s->value) {
-			nft_print(octx, "%s ", s->identifier);
-			break;
-		}
-	}
+	if (dirstr)
+		nft_print(octx, "%s ", dirstr);
 
 	switch (key) {
 	case NFT_CT_SRC:
@@ -338,26 +365,26 @@ static void ct_expr_pctx_update(struct proto_ctx *ctx, const struct expr *expr)
 	proto_ctx_update(ctx, left->ct.base + 1, &expr->location, desc);
 }
 
-static const struct expr_ops ct_expr_ops = {
+const struct expr_ops ct_expr_ops = {
 	.type		= EXPR_CT,
 	.name		= "ct",
 	.print		= ct_expr_print,
+	.json		= ct_expr_json,
 	.cmp		= ct_expr_cmp,
 	.clone		= ct_expr_clone,
 	.pctx_update	= ct_expr_pctx_update,
 };
 
 struct expr *ct_expr_alloc(const struct location *loc, enum nft_ct_keys key,
-			   int8_t direction, uint8_t nfproto)
+			   int8_t direction)
 {
 	const struct ct_template *tmpl = &ct_templates[key];
 	struct expr *expr;
 
-	expr = expr_alloc(loc, &ct_expr_ops, tmpl->dtype,
+	expr = expr_alloc(loc, EXPR_CT, tmpl->dtype,
 			  tmpl->byteorder, tmpl->len);
 	expr->ct.key = key;
 	expr->ct.direction = direction;
-	expr->ct.nfproto = nfproto;
 
 	switch (key) {
 	case NFT_CT_SRC:
@@ -393,10 +420,10 @@ void ct_expr_update_type(struct proto_ctx *ctx, struct expr *expr)
 	case NFT_CT_SRC:
 	case NFT_CT_DST:
 		if (desc == &proto_ip) {
-			expr->dtype = &ipaddr_type;
+			datatype_set(expr, &ipaddr_type);
 			expr->ct.nfproto = NFPROTO_IPV4;
 		} else if (desc == &proto_ip6) {
-			expr->dtype = &ip6addr_type;
+			datatype_set(expr, &ip6addr_type);
 			expr->ct.nfproto = NFPROTO_IPV6;
 		}
 
@@ -406,7 +433,17 @@ void ct_expr_update_type(struct proto_ctx *ctx, struct expr *expr)
 	case NFT_CT_PROTO_DST:
 		if (desc == NULL)
 			break;
-		expr->dtype = &inet_service_type;
+		datatype_set(expr, &inet_service_type);
+		break;
+	case NFT_CT_SRC_IP:
+	case NFT_CT_DST_IP:
+		expr->dtype = &ipaddr_type;
+		expr->len = expr->dtype->size;
+		break;
+	case NFT_CT_SRC_IP6:
+	case NFT_CT_DST_IP6:
+		expr->dtype = &ip6addr_type;
+		expr->len = expr->dtype->size;
 		break;
 	default:
 		break;
@@ -420,10 +457,17 @@ static void ct_stmt_print(const struct stmt *stmt, struct output_ctx *octx)
 	expr_print(stmt->ct.expr, octx);
 }
 
+static void ct_stmt_destroy(struct stmt *stmt)
+{
+	expr_free(stmt->ct.expr);
+}
+
 static const struct stmt_ops ct_stmt_ops = {
 	.type		= STMT_CT,
 	.name		= "ct",
 	.print		= ct_stmt_print,
+	.json		= ct_stmt_json,
+	.destroy	= ct_stmt_destroy,
 };
 
 struct stmt *ct_stmt_alloc(const struct location *loc, enum nft_ct_keys key,
@@ -449,6 +493,7 @@ static const struct stmt_ops notrack_stmt_ops = {
 	.type		= STMT_NOTRACK,
 	.name		= "notrack",
 	.print		= notrack_stmt_print,
+	.json		= notrack_stmt_json,
 };
 
 struct stmt *notrack_stmt_alloc(const struct location *loc)
@@ -459,13 +504,19 @@ struct stmt *notrack_stmt_alloc(const struct location *loc)
 static void flow_offload_stmt_print(const struct stmt *stmt,
 				    struct output_ctx *octx)
 {
-	printf("flow offload @%s", stmt->flow.table_name);
+	nft_print(octx, "flow add @%s", stmt->flow.table_name);
+}
+
+static void flow_offload_stmt_destroy(struct stmt *stmt)
+{
+	xfree(stmt->flow.table_name);
 }
 
 static const struct stmt_ops flow_offload_stmt_ops = {
 	.type		= STMT_FLOW_OFFLOAD,
 	.name		= "flow_offload",
 	.print		= flow_offload_stmt_print,
+	.destroy	= flow_offload_stmt_destroy,
 };
 
 struct stmt *flow_offload_stmt_alloc(const struct location *loc,
