@@ -35,6 +35,7 @@
 #include <utils.h>
 #include <erec.h>
 #include <iface.h>
+#include <json.h>
 
 static struct symbol_table *realm_tbl;
 void realm_table_meta_init(void)
@@ -206,7 +207,7 @@ static void uid_type_print(const struct expr *expr, struct output_ctx *octx)
 {
 	struct passwd *pw;
 
-	if (octx->numeric < NFT_NUMERIC_ALL) {
+	if (nft_output_guid(octx)) {
 		uint32_t uid = mpz_get_uint32(expr->value);
 
 		pw = getpwuid(uid);
@@ -251,6 +252,7 @@ const struct datatype uid_type = {
 	.size		= sizeof(uid_t) * BITS_PER_BYTE,
 	.basetype	= &integer_type,
 	.print		= uid_type_print,
+	.json		= uid_type_json,
 	.parse		= uid_type_parse,
 };
 
@@ -258,7 +260,7 @@ static void gid_type_print(const struct expr *expr, struct output_ctx *octx)
 {
 	struct group *gr;
 
-	if (octx->numeric < NFT_NUMERIC_ALL) {
+	if (nft_output_guid(octx)) {
 		uint32_t gid = mpz_get_uint32(expr->value);
 
 		gr = getgrgid(gid);
@@ -303,6 +305,7 @@ const struct datatype gid_type = {
 	.size		= sizeof(gid_t) * BITS_PER_BYTE,
 	.basetype	= &integer_type,
 	.print		= gid_type_print,
+	.json		= gid_type_json,
 	.parse		= gid_type_parse,
 };
 
@@ -334,7 +337,8 @@ const struct datatype pkttype_type = {
 	.sym_tbl	= &pkttype_type_tbl,
 };
 
-static struct symbol_table *devgroup_tbl;
+struct symbol_table *devgroup_tbl = NULL;
+
 void devgroup_table_init(void)
 {
 	devgroup_tbl = rt_symbol_table_init("/etc/iproute2/group");
@@ -365,6 +369,7 @@ const struct datatype devgroup_type = {
 	.size		= 4 * BITS_PER_BYTE,
 	.basetype	= &integer_type,
 	.print		= devgroup_type_print,
+	.json		= devgroup_type_json,
 	.parse		= devgroup_type_parse,
 	.flags		= DTYPE_F_PREFIX,
 };
@@ -378,7 +383,7 @@ const struct datatype ifname_type = {
 	.basetype	= &string_type,
 };
 
-static const struct meta_template meta_templates[] = {
+const struct meta_template meta_templates[] = {
 	[NFT_META_LEN]		= META_TEMPLATE("length",    &integer_type,
 						4 * 8, BYTEORDER_HOST_ENDIAN),
 	[NFT_META_PROTOCOL]	= META_TEMPLATE("protocol",  &ethertype_type,
@@ -437,22 +442,25 @@ static const struct meta_template meta_templates[] = {
 	[NFT_META_PRANDOM]	= META_TEMPLATE("random",    &integer_type,
 						4 * BITS_PER_BYTE,
 						BYTEORDER_BIG_ENDIAN), /* avoid conversion; doesn't have endianess */
-	[NFT_META_SECPATH]	= META_TEMPLATE("secpath", &boolean_type,
+	[NFT_META_SECPATH]	= META_TEMPLATE("ipsec", &boolean_type,
 						BITS_PER_BYTE, BYTEORDER_HOST_ENDIAN),
+	[NFT_META_IIFKIND]	= META_TEMPLATE("iifkind",   &ifname_type,
+						IFNAMSIZ * BITS_PER_BYTE,
+						BYTEORDER_HOST_ENDIAN),
+	[NFT_META_OIFKIND]	= META_TEMPLATE("oifkind",   &ifname_type,
+						IFNAMSIZ * BITS_PER_BYTE,
+						BYTEORDER_HOST_ENDIAN),
 };
 
-static bool meta_key_is_qualified(enum nft_meta_keys key)
+static bool meta_key_is_unqualified(enum nft_meta_keys key)
 {
 	switch (key) {
-	case NFT_META_LEN:
-	case NFT_META_NFPROTO:
-	case NFT_META_L4PROTO:
-	case NFT_META_PROTOCOL:
-	case NFT_META_PRIORITY:
-	case NFT_META_PRANDOM:
-	case NFT_META_SECPATH:
-	case NFT_META_BRI_IIFNAME:
-	case NFT_META_BRI_OIFNAME:
+	case NFT_META_IIF:
+	case NFT_META_OIF:
+	case NFT_META_IIFNAME:
+	case NFT_META_OIFNAME:
+	case NFT_META_IIFGROUP:
+	case NFT_META_OIFGROUP:
 		return true;
 	default:
 		return false;
@@ -461,11 +469,11 @@ static bool meta_key_is_qualified(enum nft_meta_keys key)
 
 static void meta_expr_print(const struct expr *expr, struct output_ctx *octx)
 {
-	if (meta_key_is_qualified(expr->meta.key))
-		nft_print(octx, "meta %s",
+	if (meta_key_is_unqualified(expr->meta.key))
+		nft_print(octx, "%s",
 			  meta_templates[expr->meta.key].token);
 	else
-		nft_print(octx, "%s",
+		nft_print(octx, "meta %s",
 			  meta_templates[expr->meta.key].token);
 }
 
@@ -531,7 +539,11 @@ static void meta_expr_pctx_update(struct proto_ctx *ctx,
 		proto_ctx_update(ctx, PROTO_BASE_TRANSPORT_HDR, &expr->location, desc);
 		break;
 	case NFT_META_PROTOCOL:
-		if (h->base < PROTO_BASE_NETWORK_HDR && ctx->family != NFPROTO_NETDEV)
+		if (h->base != PROTO_BASE_LL_HDR)
+			return;
+
+		if (ctx->family != NFPROTO_NETDEV &&
+		    ctx->family != NFPROTO_BRIDGE)
 			return;
 
 		desc = proto_find_upper(h->desc, ntohs(mpz_get_uint16(right->value)));
@@ -545,10 +557,11 @@ static void meta_expr_pctx_update(struct proto_ctx *ctx,
 	}
 }
 
-static const struct expr_ops meta_expr_ops = {
+const struct expr_ops meta_expr_ops = {
 	.type		= EXPR_META,
 	.name		= "meta",
 	.print		= meta_expr_print,
+	.json		= meta_expr_json,
 	.cmp		= meta_expr_cmp,
 	.clone		= meta_expr_clone,
 	.pctx_update	= meta_expr_pctx_update,
@@ -559,7 +572,7 @@ struct expr *meta_expr_alloc(const struct location *loc, enum nft_meta_keys key)
 	const struct meta_template *tmpl = &meta_templates[key];
 	struct expr *expr;
 
-	expr = expr_alloc(loc, &meta_expr_ops, tmpl->dtype,
+	expr = expr_alloc(loc, EXPR_META, tmpl->dtype,
 			  tmpl->byteorder, tmpl->len);
 	expr->meta.key = key;
 
@@ -588,20 +601,27 @@ struct expr *meta_expr_alloc(const struct location *loc, enum nft_meta_keys key)
 
 static void meta_stmt_print(const struct stmt *stmt, struct output_ctx *octx)
 {
-	if (meta_key_is_qualified(stmt->meta.key))
-		nft_print(octx, "meta %s set ",
+	if (meta_key_is_unqualified(stmt->meta.key))
+		nft_print(octx, "%s set ",
 			  meta_templates[stmt->meta.key].token);
 	else
-		nft_print(octx, "%s set ",
+		nft_print(octx, "meta %s set ",
 			  meta_templates[stmt->meta.key].token);
 
 	expr_print(stmt->meta.expr, octx);
 }
 
+static void meta_stmt_destroy(struct stmt *stmt)
+{
+	expr_free(stmt->meta.expr);
+}
+
 static const struct stmt_ops meta_stmt_ops = {
 	.type		= STMT_META,
 	.name		= "meta",
 	.print		= meta_stmt_print,
+	.json		= meta_stmt_json,
+	.destroy	= meta_stmt_destroy,
 };
 
 struct stmt *meta_stmt_alloc(const struct location *loc, enum nft_meta_keys key,
@@ -661,6 +681,9 @@ struct error_record *meta_key_parse(const struct location *loc,
 	} else if (strcmp(str, "obriport") == 0) {
 		*value = NFT_META_BRI_OIFNAME;
 		return NULL;
+	} else if (strcmp(str, "secpath") == 0) {
+		*value = NFT_META_SECPATH;
+		return NULL;
 	}
 
 	len = (int)sizeof(buf);
