@@ -1,4 +1,4 @@
-#!/usr/bin/python
+#!/usr/bin/python2
 #
 # (C) 2014 by Ana Rey Botello <anarey@gmail.com>
 #
@@ -17,8 +17,14 @@ import sys
 import os
 import argparse
 import signal
+import json
+import traceback
 
 TESTS_PATH = os.path.dirname(os.path.abspath(__file__))
+sys.path.insert(0, os.path.join(TESTS_PATH, '../../py/'))
+
+from nftables import Nftables
+
 TESTS_DIRECTORY = ["any", "arp", "bridge", "inet", "ip", "ip6"]
 LOGFILE = "/tmp/nftables-test.log"
 log_file = None
@@ -103,35 +109,67 @@ class Obj:
         return self.__dict__ == other.__dict__
 
 
-def print_msg(reason, filename=None, lineno=None, color=None, errstr=None):
+def print_msg(reason, errstr, filename=None, lineno=None, color=None):
     '''
     Prints a message with nice colors, indicating file and line number.
     '''
     if filename and lineno:
-        print filename + ": " + color + "ERROR:" + Colors.ENDC + \
-              " line %d: %s" % (lineno + 1, reason)
+        sys.stderr.write(filename + ": " + color + errstr + Colors.ENDC + \
+              " line %d: %s" % (lineno + 1, reason))
     else:
-        print color + "ERROR:" + Colors.ENDC + " %s" % reason
+        sys.stderr.write(color + errstr + Colors.ENDC + " %s" % reason)
+    sys.stderr.write("\n")
+    sys.stderr.flush() # So that the message stay in the right place.
 
 
 def print_error(reason, filename=None, lineno=None):
-    print_msg(reason, filename, lineno, Colors.RED, "ERROR:")
+    print_msg(reason, "ERROR:", filename, lineno, Colors.RED)
 
 
 def print_warning(reason, filename=None, lineno=None):
-    print_msg(reason, filename, lineno, Colors.YELLOW, "WARNING:")
+    print_msg(reason, "WARNING:", filename, lineno, Colors.YELLOW)
 
+def print_info(reason, filename=None, lineno=None):
+    print_msg(reason, "INFO:", filename, lineno, Colors.GREEN)
+
+def color_differences(rule, other, color):
+    rlen = len(rule)
+    olen = len(other)
+    out = ""
+
+    # find equal part at start
+    for i in range(rlen):
+        if i >= olen or rule[i] != other[i]:
+            break
+    if i > 0:
+        out += rule[:i]
+        rule = rule[i:]
+        other = other[i:]
+        rlen = len(rule)
+        olen = len(other)
+
+    # find equal part at end
+    for i in range(1, rlen + 1):
+        if i > olen or rule[rlen - i] != other[olen - i]:
+            i -= 1
+            break
+    if rlen > i:
+        out += color + rule[:rlen - i] + Colors.ENDC
+        rule = rule[rlen - i:]
+
+    out += rule
+    return out
 
 def print_differences_warning(filename, lineno, rule1, rule2, cmd):
-    reason = "'" + rule1 + "' mismatches '" + rule2 + "'"
-    print filename + ": " + Colors.YELLOW + "WARNING: " + Colors.ENDC + \
-          "line: " + str(lineno + 1) + ": '" + cmd + "': " + reason
+    colored_rule1 = color_differences(rule1, rule2, Colors.YELLOW)
+    colored_rule2 = color_differences(rule2, rule1, Colors.YELLOW)
+    reason = "'%s': '%s' mismatches '%s'" % (cmd, colored_rule1, colored_rule2)
+    print_warning(reason, filename, lineno)
 
 
 def print_differences_error(filename, lineno, cmd):
-    reason = "Listing is broken."
-    print filename + ": " + Colors.RED + "ERROR: " + Colors.ENDC + "line: " + \
-          str(lineno + 1) + ": '" + cmd + "': " + reason
+    reason = "'%s': Listing is broken." % cmd
+    print_error(reason, filename, lineno)
 
 
 def table_exist(table, filename, lineno):
@@ -139,7 +177,7 @@ def table_exist(table, filename, lineno):
     Exists a table.
     '''
     cmd = "list table %s" % table
-    ret = execute_cmd(cmd, filename, lineno, numeric="all")
+    ret = execute_cmd(cmd, filename, lineno)
 
     return True if (ret == 0) else False
 
@@ -226,7 +264,7 @@ def chain_exist(chain, table, filename):
     Checks a chain
     '''
     cmd = "list chain %s %s" % (table, chain)
-    ret = execute_cmd(cmd, filename, chain.lineno, numeric="all")
+    ret = execute_cmd(cmd, filename, chain.lineno)
 
     return True if (ret == 0) else False
 
@@ -241,7 +279,9 @@ def chain_create(chain, table, filename):
         print_error(reason, filename, chain.lineno)
         return -1
 
-    cmd = "add chain %s %s { %s; }" % (table, chain, chain.config)
+    cmd = "add chain %s %s" % (table, chain)
+    if chain.config:
+        cmd += " { %s; }" % chain.config
 
     ret = execute_cmd(cmd, filename, chain.lineno)
     if ret != 0:
@@ -271,14 +311,14 @@ def chain_delete(chain, table, filename=None, lineno=None):
     cmd = "flush chain %s %s" % (table, chain)
     ret = execute_cmd(cmd, filename, lineno)
     if ret != 0:
-        reason = "I cannot flush this chain " + chain
+        reason = "I cannot " + cmd
         print_error(reason, filename, lineno)
         return -1
 
     cmd = "delete chain %s %s" % (table, chain)
     ret = execute_cmd(cmd, filename, lineno)
     if ret != 0:
-        reason = cmd + "I cannot delete this chain " + chain
+        reason = "I cannot " + cmd
         print_error(reason, filename, lineno)
         return -1
 
@@ -427,7 +467,7 @@ def set_exist(set_name, table, filename, lineno):
     Check if the set exists.
     '''
     cmd = "list set %s %s" % (table, set_name)
-    ret = execute_cmd(cmd, filename, lineno, numeric="all")
+    ret = execute_cmd(cmd, filename, lineno)
 
     return True if (ret == 0) else False
 
@@ -437,7 +477,7 @@ def _set_exist(s, filename, lineno):
     Check if the set exists.
     '''
     cmd = "list set %s %s %s" % (s.family, s.table, s.name)
-    ret = execute_cmd(cmd, filename, lineno, numeric="all")
+    ret = execute_cmd(cmd, filename, lineno)
 
     return True if (ret == 0) else False
 
@@ -545,7 +585,7 @@ def obj_exist(o, table, filename, lineno):
     Check if the object exists.
     '''
     cmd = "list %s %s %s" % (o.type, table, o.name)
-    ret = execute_cmd(cmd, filename, lineno, numeric="all")
+    ret = execute_cmd(cmd, filename, lineno)
 
     return True if (ret == 0) else False
 
@@ -555,7 +595,7 @@ def _obj_exist(o, filename, lineno):
     Check if the object exists.
     '''
     cmd = "list %s %s %s %s" % (o.type, o.family, o.table, o.name)
-    ret = execute_cmd(cmd, filename, lineno, numeric="all")
+    ret = execute_cmd(cmd, filename, lineno)
 
     return True if (ret == 0) else False
 
@@ -614,6 +654,9 @@ def payload_check(payload_buffer, file, cmd):
     file.seek(0, 0)
     i = 0
 
+    if not payload_buffer:
+        return False
+
     for lineno, want_line in enumerate(payload_buffer):
         line = file.readline()
 
@@ -636,6 +679,23 @@ def payload_check(payload_buffer, file, cmd):
     return i > 0
 
 
+def json_dump_normalize(json_string, human_readable = False):
+    json_obj = json.loads(json_string)
+
+    if human_readable:
+        return json.dumps(json_obj, sort_keys = True,
+                          indent = 4, separators = (',', ': '))
+    else:
+        return json.dumps(json_obj, sort_keys = True)
+
+def json_validate(json_string):
+    json_obj = json.loads(json_string)
+    try:
+        nftables.json_validate(json_obj)
+    except Exception:
+        print_error("schema validation failed for input '%s'" % json_string)
+        print_error(traceback.format_exc())
+
 def rule_add(rule, filename, lineno, force_all_family_option, filename_path):
     '''
     Adds a rule
@@ -648,23 +708,62 @@ def rule_add(rule, filename, lineno, force_all_family_option, filename_path):
         print_error(reason, filename, lineno)
         return [-1, warning, error, unit_tests]
 
-    payload_expected = []
+    if rule[1].strip() == "ok":
+        payload_expected = None
+        try:
+            payload_log = open("%s.payload" % filename_path)
+            payload_expected = payload_find_expected(payload_log, rule[0])
+        except:
+            payload_log = None
+
+        if enable_json_option:
+            try:
+                json_log = open("%s.json" % filename_path)
+                json_input = json_find_expected(json_log, rule[0])
+            except:
+                json_input = None
+
+            if not json_input:
+                print_error("did not find JSON equivalent for rule '%s'"
+                            % rule[0])
+            else:
+                try:
+                    json_input = json_dump_normalize(json_input)
+                except ValueError:
+                    reason = "Invalid JSON syntax in rule: %s" % json_input
+                    print_error(reason)
+                    return [-1, warning, error, unit_tests]
+
+            try:
+                json_log = open("%s.json.output" % filename_path)
+                json_expected = json_find_expected(json_log, rule[0])
+            except:
+                # will use json_input for comparison
+                json_expected = None
+
+            if json_expected:
+                try:
+                    json_expected = json_dump_normalize(json_expected)
+                except ValueError:
+                    reason = "Invalid JSON syntax in expected output: %s" % json_expected
+                    print_error(reason)
+                    return [-1, warning, error, unit_tests]
 
     for table in table_list:
-        try:
-            payload_log = open("%s.payload.%s" % (filename_path, table.family))
-        except IOError:
-            payload_log = open("%s.payload" % filename_path)
-
         if rule[1].strip() == "ok":
+            table_payload_expected = None
             try:
-                payload_expected.index(rule[0])
-            except ValueError:
-                payload_expected = payload_find_expected(payload_log, rule[0])
-
-                if not payload_expected:
+                payload_log = open("%s.payload.%s" % (filename_path, table.family))
+                table_payload_expected = payload_find_expected(payload_log, rule[0])
+            except:
+                if not payload_log:
+                    print_error("did not find any payload information",
+                                filename_path)
+                elif not payload_expected:
                     print_error("did not find payload information for "
                                 "rule '%s'" % rule[0], payload_log.name, 1)
+            if not table_payload_expected:
+                table_payload_expected = payload_expected
 
         for table_chain in table.chains:
             chain = chain_get_by_name(table_chain)
@@ -673,6 +772,7 @@ def rule_add(rule, filename, lineno, force_all_family_option, filename_path):
 
             payload_log = os.tmpfile()
 
+            # Add rule and check return code
             cmd = "add rule %s %s %s" % (table, chain, rule[0])
             ret = execute_cmd(cmd, filename, lineno, payload_log, debug="netlink")
 
@@ -693,77 +793,200 @@ def rule_add(rule, filename, lineno, force_all_family_option, filename_path):
                 ret = 0
                 continue
 
-            if ret == 0:
-                # Check for matching payload
-                if state == "ok" and not payload_check(payload_expected,
-                                                       payload_log, cmd):
-                    error += 1
-                    gotf = open("%s.payload.got" % filename_path, 'a')
-                    payload_log.seek(0, 0)
-                    gotf.write("# %s\n" % rule[0])
-                    while True:
-                        line = payload_log.readline()
-                        if line == "":
-                            break
-                        gotf.write(line)
-                    gotf.close()
-                    print_warning("Wrote payload for rule %s" % rule[0],
-                                  gotf.name, 1)
+            if ret != 0:
+                continue
 
-                # Check output of nft
-                numeric_old = nftables.set_numeric_output("all")
-                stateless_old = nftables.set_stateless_output(True)
-                list_cmd = 'list table %s' % table
-                rc, pre_output, err = nftables.cmd(list_cmd)
-                nftables.set_numeric_output(numeric_old)
-                nftables.set_stateless_output(stateless_old)
+            # Check for matching payload
+            if state == "ok" and not payload_check(table_payload_expected,
+                                                   payload_log, cmd):
+                error += 1
+                gotf = open("%s.payload.got" % filename_path, 'a')
+                payload_log.seek(0, 0)
+                gotf.write("# %s\n" % rule[0])
+                while True:
+                    line = payload_log.readline()
+                    if line == "":
+                        break
+                    gotf.write(line)
+                gotf.close()
+                print_warning("Wrote payload for rule %s" % rule[0],
+                              gotf.name, 1)
 
-                output = pre_output.split(";")
-                if len(output) < 2:
-                    reason = cmd + ": Listing is broken."
-                    print_error(reason, filename, lineno)
+            # Check for matching ruleset listing
+            numeric_proto_old = nftables.set_numeric_proto_output(True)
+            stateless_old = nftables.set_stateless_output(True)
+            list_cmd = 'list table %s' % table
+            rc, pre_output, err = nftables.cmd(list_cmd)
+            nftables.set_numeric_proto_output(numeric_proto_old)
+            nftables.set_stateless_output(stateless_old)
+
+            output = pre_output.split(";")
+            if len(output) < 2:
+                reason = cmd + ": Listing is broken."
+                print_error(reason, filename, lineno)
+                ret = -1
+                error += 1
+                if not force_all_family_option:
+                    return [ret, warning, error, unit_tests]
+                continue
+
+            rule_output = output_clean(pre_output, chain)
+            retest_output = False
+            if len(rule) == 3:
+                teoric_exit = rule[2]
+                retest_output = True
+            else:
+                teoric_exit = rule[0]
+
+            if rule_output.rstrip() != teoric_exit.rstrip():
+                if rule[0].find("{") != -1:  # anonymous sets
+                    if set_check_element(teoric_exit.rstrip(),
+                                         rule_output.rstrip()) != 0:
+                        warning += 1
+                        retest_output = True
+                        print_differences_warning(filename, lineno,
+                                                  teoric_exit.rstrip(),
+                                                  rule_output, cmd)
+                        if not force_all_family_option:
+                            return [ret, warning, error, unit_tests]
+                else:
+                    if len(rule_output) <= 0:
+                        error += 1
+                        print_differences_error(filename, lineno, cmd)
+                        if not force_all_family_option:
+                            return [ret, warning, error, unit_tests]
+
+                    warning += 1
+                    retest_output = True
+                    print_differences_warning(filename, lineno,
+                                              teoric_exit.rstrip(),
+                                              rule_output, cmd)
+
+                    if not force_all_family_option:
+                        return [ret, warning, error, unit_tests]
+
+            if retest_output:
+                table_flush(table, filename, lineno)
+
+                # Add rule and check return code
+                cmd = "add rule %s %s %s" % (table, chain, rule_output.rstrip())
+                ret = execute_cmd(cmd, filename, lineno, payload_log, debug="netlink")
+
+                if ret != 0:
+                    test_state = "Replaying rule failed."
+                    reason = cmd + ": " + test_state
+                    print_warning(reason, filename, lineno)
                     ret = -1
                     error += 1
                     if not force_all_family_option:
                         return [ret, warning, error, unit_tests]
-                else:
-                    rule_output = output_clean(pre_output, chain)
-                    if len(rule) == 3:
-                        teoric_exit = rule[2]
-                    else:
-                        teoric_exit = rule[0]
+                # Check for matching payload
+                elif not payload_check(table_payload_expected,
+                                       payload_log, cmd):
+                    error += 1
 
-                    if rule_output.rstrip() != teoric_exit.rstrip():
-                        if rule[0].find("{") != -1:  # anonymous sets
-                            if set_check_element(teoric_exit.rstrip(), rule_output.rstrip()) != 0:
-                                warning += 1
-                                print_differences_warning(filename, lineno,
-                                                          teoric_exit.rstrip(),
-                                                          rule_output, cmd)
-                                if not force_all_family_option:
-                                    return [ret, warning, error, unit_tests]
-                        else:
-                            if len(rule_output) <= 0:
-                                error += 1
-                                print_differences_error(filename, lineno, cmd)
-                                if not force_all_family_option:
-                                    return [ret, warning, error, unit_tests]
+            if not enable_json_option:
+                continue
 
-                            warning += 1
-                            print_differences_warning(filename, lineno,
-                                                      teoric_exit.rstrip(),
-                                                      rule_output, cmd)
+            # Generate JSON equivalent for rule if not found
+            if not json_input:
+                json_old = nftables.set_json_output(True)
+                rc, json_output, err = nftables.cmd(list_cmd)
+                nftables.set_json_output(json_old)
 
-                            if not force_all_family_option:
-                                return [ret, warning, error, unit_tests]
+                json_output = json.loads(json_output)
+                for item in json_output["nftables"]:
+                    if "rule" in item:
+                        del(item["rule"]["handle"])
+                        json_output = item["rule"]
+                        break
+                json_input = json.dumps(json_output["expr"], sort_keys = True)
+
+                gotf = open("%s.json.got" % filename_path, 'a')
+                jdump = json_dump_normalize(json_input, True)
+                gotf.write("# %s\n%s\n\n" % (rule[0], jdump))
+                gotf.close()
+                print_warning("Wrote JSON equivalent for rule %s" % rule[0],
+                              gotf.name, 1)
+
+            table_flush(table, filename, lineno)
+            payload_log = os.tmpfile()
+
+            # Add rule in JSON format
+            cmd = json.dumps({ "nftables": [{ "add": { "rule": {
+                    "family": table.family,
+                    "table": table.name,
+                    "chain": chain.name,
+                    "expr": json.loads(json_input),
+            }}}]})
+
+            if enable_json_schema:
+                json_validate(cmd)
+
+            json_old = nftables.set_json_output(True)
+            ret = execute_cmd(cmd, filename, lineno, payload_log, debug="netlink")
+            nftables.set_json_output(json_old)
+
+            if ret != 0:
+                reason = "Failed to add JSON equivalent rule"
+                print_error(reason, filename, lineno)
+                continue
+
+            # Check for matching payload
+            if not payload_check(table_payload_expected, payload_log, cmd):
+                error += 1
+                gotf = open("%s.json.payload.got" % filename_path, 'a')
+                payload_log.seek(0, 0)
+                gotf.write("# %s\n" % rule[0])
+                while True:
+                    line = payload_log.readline()
+                    if line == "":
+                        break
+                    gotf.write(line)
+                gotf.close()
+                print_warning("Wrote JSON payload for rule %s" % rule[0],
+                              gotf.name, 1)
+
+            # Check for matching ruleset listing
+            numeric_proto_old = nftables.set_numeric_proto_output(True)
+            stateless_old = nftables.set_stateless_output(True)
+            json_old = nftables.set_json_output(True)
+            rc, json_output, err = nftables.cmd(list_cmd)
+            nftables.set_json_output(json_old)
+            nftables.set_numeric_proto_output(numeric_proto_old)
+            nftables.set_stateless_output(stateless_old)
+
+            if enable_json_schema:
+                json_validate(json_output)
+
+            json_output = json.loads(json_output)
+            for item in json_output["nftables"]:
+                if "rule" in item:
+                    del(item["rule"]["handle"])
+                    json_output = item["rule"]
+                    break
+            json_output = json.dumps(json_output["expr"], sort_keys = True)
+
+            if not json_expected and json_output != json_input:
+                print_differences_warning(filename, lineno,
+                                          json_input, json_output, cmd)
+                error += 1
+                gotf = open("%s.json.output.got" % filename_path, 'a')
+                jdump = json_dump_normalize(json_output, True)
+                gotf.write("# %s\n%s\n\n" % (rule[0], jdump))
+                gotf.close()
+                print_warning("Wrote JSON output for rule %s" % rule[0],
+                              gotf.name, 1)
+                # prevent further warnings and .got file updates
+                json_expected = json_output
+            elif json_expected and json_output != json_expected:
+                print_differences_warning(filename, lineno,
+                                          json_expected, json_output, cmd)
+                error += 1
 
     return [ret, warning, error, unit_tests]
 
 
-def preexec():
-    os.setpgrp()  # Don't forward signals.
-
-
 def cleanup_on_exit():
     for table in table_list:
         for table_chain in table.chains:
@@ -781,8 +1004,7 @@ def signal_handler(signal, frame):
     signal_received = 1
 
 
-def execute_cmd(cmd, filename, lineno,
-                stdout_log=False, numeric=False, debug=False):
+def execute_cmd(cmd, filename, lineno, stdout_log=False, debug=False):
     '''
     Executes a command, checks for segfaults and returns the command exit
     code.
@@ -791,7 +1013,6 @@ def execute_cmd(cmd, filename, lineno,
     :param filename: name of the file tested (used for print_error purposes)
     :param lineno: line number being tested (used for print_error purposes)
     :param stdout_log: redirect stdout to this file instead of global log_file
-    :param numeric: turn numeric output temporarily on
     :param debug: temporarily set these debug flags
     '''
     global log_file
@@ -799,9 +1020,6 @@ def execute_cmd(cmd, filename, lineno,
     if debug_option:
         print cmd
 
-    if numeric:
-        numeric_old = nftables.get_numeric_output()
-        nftables.set_numeric_output(numeric)
     if debug:
         debug_old = nftables.get_debug()
         nftables.set_debug(debug)
@@ -816,8 +1034,6 @@ def execute_cmd(cmd, filename, lineno,
     log_file.write(err)
     log_file.flush()
 
-    if numeric:
-        nftables.set_numeric_output(numeric_old)
     if debug:
         nftables.set_debug(debug_old)
 
@@ -903,6 +1119,10 @@ def obj_process(obj_line, filename, lineno):
        obj_type = "ct helper"
        tokens[3] = ""
 
+    if obj_type == "ct" and tokens[3] == "timeout":
+       obj_type = "ct timeout"
+       tokens[3] = ""
+
     if len(tokens) > 3:
         obj_spcf = " ".join(tokens[3:])
 
@@ -947,6 +1167,37 @@ def payload_find_expected(payload_log, rule):
     return payload_buffer
 
 
+def json_find_expected(json_log, rule):
+    '''
+    Find the corresponding JSON for given rule
+
+    :param json_log: open file handle of the json data
+    :param rule: nft rule we are going to add
+    '''
+    found = 0
+    json_buffer = ""
+
+    while True:
+        line = json_log.readline()
+        if not line:
+            break
+
+        if line[0] == "#":  # rule start
+            rule_line = line.strip()[2:]
+
+            if rule_line == rule.strip():
+                found = 1
+                continue
+
+        if found == 1:
+            json_buffer += line.rstrip("\n").strip()
+            if line.isspace():
+                return json_buffer
+
+    json_log.seek(0, 0)
+    return json_buffer
+
+
 def run_test_file(filename, force_all_family_option, specific_file):
     '''
     Runs a test file
@@ -1100,11 +1351,21 @@ def main():
                         dest='force_all_family',
                         help='keep testing all families on error')
 
+    parser.add_argument('-j', '--enable-json', action='store_true',
+                        dest='enable_json',
+                        help='test JSON functionality as well')
+
+    parser.add_argument('-s', '--schema', action='store_true',
+                        dest='enable_schema',
+                        help='verify json input/output against schema')
+
     args = parser.parse_args()
-    global debug_option, need_fix_option
+    global debug_option, need_fix_option, enable_json_option, enable_json_schema
     debug_option = args.debug
     need_fix_option = args.need_fix_line
     force_all_family_option = args.force_all_family
+    enable_json_option = args.enable_json
+    enable_json_schema = args.enable_schema
     specific_file = False
 
     signal.signal(signal.SIGINT, signal_handler)
@@ -1117,24 +1378,26 @@ def main():
     # Change working directory to repository root
     os.chdir(TESTS_PATH + "/../..")
 
-    sys.path.append('py/')
-    from nftables import Nftables
-
     if not os.path.exists('src/.libs/libnftables.so'):
         print "The nftables library does not exist. " \
               "You need to build the project."
         return
 
+    if args.enable_schema and not args.enable_json:
+        print_error("Option --schema requires option --json")
+        return
+
     global nftables
-    nftables = Nftables('src/.libs/libnftables.so')
+    nftables = Nftables(sofile = 'src/.libs/libnftables.so')
 
     test_files = files_ok = run_total = 0
     tests = passed = warnings = errors = 0
     global log_file
     try:
         log_file = open(LOGFILE, 'w')
+        print_info("Log will be available at %s" % LOGFILE)
     except IOError:
-        print "Cannot open log file %s" % LOGFILE
+        print_error("Cannot open log file %s" % LOGFILE)
         return
 
     file_list = []
